#include <exec/types.h>
#include <exec/memory.h>
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/layers.h>
#include <proto/intuition.h>
#include <proto/dos.h>
#include <proto/cybergraphics.h>
#include <clib/macros.h>
#include <dos/dos.h>
#include <dos/datetime.h>
#include <intuition/intuitionbase.h>
#include <intuition/gadgetclass.h>
#include <intuition/imageclass.h>
#include <intuition/preferences.h>
#include <graphics/layers.h>
#include <graphics/rpattr.h>
#include <graphics/gfxmacros.h>
#include <cybergraphx/cybergraphics.h>

#include "intuition_preferences.h"
#include "intuition_intern.h"
#include "boopsigadgets.h"
#include "intuition_customize.h"
#include "intuition_extend.h"
#include "mosmisc.h"
#include "showhide.h"

#include <string.h>


#undef DEBUG
#define DEBUG 0
#include <aros/debug.h>

ULONG addextragadget(struct Window *w,BOOL is_gzz,struct DrawInfo *dri,LONG relright,ULONG imagetype,ULONG gadgetid,ULONG gadgettype,struct IntuitionBase *IntuitionBase);
extern ULONG HookEntry();

/**********************************************************************************/

void LoadDefaultPreferences(struct IntuitionBase * IntuitionBase)
{
	static UWORD DriPens2[NUMDRIPENS] = { 1, 0, 1, 1, 1, 1, 0, 0, 1, 0, 1, 1 , 1 , 0};
	static UWORD DriPens4[NUMDRIPENS] = { 1, 0, 1, 2, 1, 3, 1, 0, 2, 1, 2, 1 , 2 , 1};
	static UWORD DriPens8[NUMDRIPENS] = { 1, 0, 1, 2, 1, 3, 1, 0, 2, 1, 2, 1 , -4 , -3};

	/*
	** Load the intuition preferences from a file on the disk
	*/
#ifdef SKINS
	memclr(&IBase->AmbientScreenMode,sizeof (struct AmbientScreenDesc));
	IBase->AmbientScreenMode.ModeID = INVALID_ID;
	IBase->AmbientScreenMode.DisplayWidth = -1;
	IBase->AmbientScreenMode.DisplayHeight = -1;
	IBase->AmbientScreenMode.DisplayDepth = 24;
	IBase->AmbientScreenMode.Compositing = TRUE;
	IBase->AmbientScreenMode.AutoScroll = TRUE;
#else
	IBase->ScreenModePrefs.smp_DisplayID = -1UL;
	IBase->ScreenModePrefs.smp_Width = AROS_DEFAULT_WBWIDTH;
	IBase->ScreenModePrefs.smp_Height = AROS_DEFAULT_WBHEIGHT;
	IBase->ScreenModePrefs.smp_Depth = AROS_DEFAULT_WBDEPTH;
	IBase->ScreenModePrefs.smp_Control = 0;
#endif

	/*
	 * Mouse default.
	 */
	IBase->DefaultPreferences.PointerTicks = 2;

	CopyMem(&IntuitionDefaultPreferences,&IBase->DefaultPreferences,sizeof(struct Preferences));
	CopyMem(&IBase->DefaultPreferences,&IBase->ActivePreferences,sizeof(struct Preferences));

	CopyMem(DriPens2, IBase->DriPens2, sizeof(DriPens2));
	CopyMem(DriPens4, IBase->DriPens4, sizeof(DriPens4));
	CopyMem(DriPens8, IBase->DriPens8, sizeof(DriPens8));
}

/**********************************************************************************/

void CheckRectFill(struct RastPort *rp, WORD x1, WORD y1, WORD x2, WORD y2,
				   struct IntuitionBase * IntuitionBase)
{

	if ((x2 >= x1) && (y2 >= y1))
	{
		RectFill(rp, x1, y1, x2, y2);
	}
}

/**********************************************************************************/

#define TITLEBAR_HEIGHT (w->BorderTop)

BOOL CreateWinSysGadgets(struct Window *w, struct IntuitionBase *IntuitionBase)
{

	struct DrawInfo *dri;
	BOOL is_gzz;

	EnterFunc(bug("CreateWinSysGadgets(w=%p)\n", w));

	is_gzz = (w->Flags & WFLG_GIMMEZEROZERO) ? TRUE : FALSE;

	dri = GetScreenDrawInfo(w->WScreen);
	if (dri)
	{
		LONG db_left, db_width, relright,ewidth; /* dragbar sizes */
		BOOL sysgads_ok = TRUE;


		db_left = 0;
		db_width = 0; /* Georg Steger: was w->Width; */

		/* Now find out what gadgets the window wants */
#if 0
		if (    w->Flags & WFLG_CLOSEGADGET
		        || w->Flags & WFLG_DEPTHGADGET
		        || w->Flags & WFLG_HASZOOM
		        || w->Flags & WFLG_DRAGBAR /* To assure w->BorderTop being set correctly */
		        || w->Flags & WFLG_SIZEGADGET
		   )
		{
			/* If any of titlebar gadgets are present, me might just as well
			insert a dragbar too */

			w->Flags |= WFLG_DRAGBAR;

			/* Georg Steger: bordertop is set by rom/intuition/openwindow.c
			to scr->WBorTop + FontHeight + 1 */

			/* w->BorderTop = TITLEBAR_HEIGHT; */
		}
#endif

		/* Relright of rightmost button */
		//relright = - (TITLEBAR_HEIGHT - 1);

		relright = 1;

		/* Now try to create the various gadgets */

		if (w->Flags & WFLG_SIZEGADGET)
		{
			/* this code must not change the 'relright' variable */
			WORD width  = ((struct IntWindow *)w)->sizeimage_width;
			WORD height = ((struct IntWindow *)w)->sizeimage_height;

			struct TagItem size_tags[] =
		    {
			    {GA_Image	, 0				},
			    {GA_RelRight	, -width  + 1	},
			    {GA_RelBottom	, -height + 1	},
			    {GA_Width	, width		},
			    {GA_Height	, height	},
			    {GA_SysGadget	, TRUE		},
			    {GA_SysGType	, GTYP_SIZING	},
			    {GA_BottomBorder, TRUE		},
			    {GA_RightBorder	, TRUE		},
			    {GA_GZZGadget	, is_gzz	},
			    {TAG_DONE			}
		    };

			struct TagItem image_tags[] =
		    {
			    {IA_Width	, width	},
			    {IA_Height	, height    },
			    {SYSIA_Which	, SIZEIMAGE },
			    {SYSIA_DrawInfo , (IPTR)dri },
			    {SYSIA_Size, w->WScreen->Flags & SCREENHIRES ? SYSISIZE_MEDRES : SYSISIZE_LOWRES},
			    {TAG_DONE		    }
		    };
			Object *im;

			im = NewObjectA(NULL, SYSICLASS, image_tags);
			if (!im)
			{
				sysgads_ok = FALSE;
			}
			else
			{
				size_tags[0].ti_Data = (IPTR)im;

				SYSGAD(w, SIZEGAD) = NewObjectA(NULL, BUTTONGCLASS, size_tags);

				if (!SYSGAD(w, SIZEGAD))
				{
					DisposeObject(im);
					sysgads_ok = FALSE;
				}
			}
		} else {
			if (w->MoreFlags & WMFLG_THINBORDERS)
			{
				WORD width  = ((struct IntWindow *)w)->sizeimage_width;
				WORD height = ((struct IntWindow *)w)->sizeimage_height;

				struct TagItem size_tags[] =
				{
				    {GA_RelRight	, -width  + 1	},
				    {GA_RelBottom	, -height + 1	},
				    {GA_Width	, width		},
				    {GA_Height	, height	},
				    {GA_SysGadget	, TRUE		},
				    {GA_SysGType	, GTYP_SIZING	},
				    {GA_BottomBorder, TRUE		},
				    {GA_RightBorder	, TRUE		},
				    {GA_GZZGadget	, is_gzz	},
				    {TAG_DONE			}
				};

				SYSGAD(w, SIZEGAD) = NewObjectA(NULL, GADGETCLASS, size_tags);

				if (!SYSGAD(w, SIZEGAD))
				{
					sysgads_ok = FALSE;
				}

				w->Flags |= WFLG_SIZEGADGET;
			}
		}

		if (w->Flags & WFLG_DEPTHGADGET)
		{
			struct TagItem depth_tags[] =
			    {
				    {GA_Image	, 0					},
				    //{GA_RelRight	, relright		},
				    {GA_Top		, 0					},
				    /*{GA_Width	, TITLEBAR_HEIGHT	},*/
				    {GA_Height	, TITLEBAR_HEIGHT	},
				    {GA_SysGadget	, TRUE			},
				    {GA_SysGType	, GTYP_WDEPTH		},
				    {GA_TopBorder	, TRUE			},
				    {GA_GZZGadget	, is_gzz		},
				    {GA_RelVerify	, TRUE			},
				    {TAG_DONE				}
			    };

			struct TagItem image_tags[] =
			    {
#ifdef SKINS
			    {TAG_IGNORE	, 0			},
#else
				{IA_Left	, -1			},
#endif
				    /*{IA_Width	, TITLEBAR_HEIGHT + 1	},*/
				    {IA_Height	, TITLEBAR_HEIGHT	},
				    {SYSIA_Which	, DEPTHIMAGE		},
				    {SYSIA_DrawInfo , (IPTR)dri		},
				    {SYSIA_Size	, w->WScreen->Flags & SCREENHIRES ?
				     SYSISIZE_MEDRES : SYSISIZE_LOWRES},
				    {TAG_DONE				}
			    };
			Object *im;

			im = NewObjectA(NULL, SYSICLASS, image_tags);
			if (!im)
			{
				sysgads_ok = FALSE;
			}
			else
			{
				depth_tags[0].ti_Data = (IPTR)im;

				SYSGAD(w, DEPTHGAD) = NewObjectA(IBase->windowsysiclass, NULL, depth_tags);

				if (!SYSGAD(w, DEPTHGAD))
				{
					DisposeObject(im);
					sysgads_ok = FALSE;
				}
				else
				{
					int width;
					GetAttr(GA_Width, SYSGAD(w, DEPTHGAD), (ULONG*)&width);

					/*****/
#ifndef SKINS
					--width;
#endif

					relright -= width;
					db_width -= width;

					{
						struct TagItem gadtags[] =
						    {
							    {GA_RelRight, relright	},
							    /* {GA_Width	, width	    }, */
							    {TAG_DONE		    }
						    };

						SetAttrsA(SYSGAD(w, DEPTHGAD), gadtags);
					}
				}
			}

		}

		if (w->Flags & WFLG_HASZOOM)
		{
			struct TagItem zoom_tags[] =
			    {
				    {GA_Image	, 0				},
				    //{GA_RelRight	, relright		},
				    {GA_Top		, 0					},
				    /*{GA_Width	, TITLEBAR_HEIGHT	},*/
				    {GA_Height	, TITLEBAR_HEIGHT	},
				    {GA_SysGadget	, TRUE			},
				    {GA_SysGType	, GTYP_WZOOM		},
				    {GA_TopBorder	, TRUE			},
				    {GA_GZZGadget	, is_gzz		},
				    {GA_RelVerify	, TRUE			},
				    {TAG_DONE				}
			    };

			struct TagItem image_tags[] =
			    {
#ifdef SKINS
			    	{TAG_IGNORE	, 0			},
#else
			    	{IA_Left	, -1			},
#endif
				    //{IA_Width	, TITLEBAR_HEIGHT + 1   },
				    {IA_Height	, TITLEBAR_HEIGHT	},
				    {SYSIA_Which	, ZOOMIMAGE		},
				    {SYSIA_DrawInfo , (IPTR)dri		},
				    {SYSIA_Size	, w->WScreen->Flags & SCREENHIRES ?
				     SYSISIZE_MEDRES : SYSISIZE_LOWRES},
				    {TAG_DONE				}
			    };
			Object *im;

			im = NewObjectA(NULL, SYSICLASS, image_tags);
			if (!im)
			{
				sysgads_ok = FALSE;
			}
			else
			{
				zoom_tags[0].ti_Data = (IPTR)im;

				SYSGAD(w, ZOOMGAD) = NewObjectA(IBase->windowsysiclass, NULL, zoom_tags);

				if (!SYSGAD(w, ZOOMGAD))
				{
					DisposeObject(im);
					sysgads_ok = FALSE;
				}
				else
				{
					int width;
					GetAttr(GA_Width, SYSGAD(w, ZOOMGAD), (ULONG*)&width);

					--width;

					relright -= width;
					db_width -= width;

					{
						struct TagItem gadtags[] =
						    {
							    {GA_RelRight, relright},
							    {TAG_DONE		    }
						    };


						SetAttrsA(SYSGAD(w, ZOOMGAD), gadtags);
					}
				}
			}
		}

		if (((struct IntWindow *)(w))->extrabuttons & ETG_LOCK)
		{
			ewidth = addextragadget(w,is_gzz,dri,relright,LOCKIMAGE,((struct IntWindow *)w)->extrabuttonsid + ETD_Lock,LOCKGAD,IntuitionBase);
			relright -= ewidth;
			db_width -= ewidth;
		}

		if (((struct IntWindow *)(w))->extrabuttons & ETG_ICONIFY)
		{
			ewidth = addextragadget(w,is_gzz,dri,relright,ICONIFYIMAGE,((struct IntWindow *)w)->extrabuttonsid + ETD_Iconify,ICONIFYGAD,IntuitionBase);
			relright -= ewidth;
			db_width -= ewidth;
		}

		if (((struct IntWindow *)(w))->extrabuttons & ETG_JUMP)
		{
			ewidth = addextragadget(w,is_gzz,dri,relright,JUMPIMAGE,((struct IntWindow *)w)->extrabuttonsid + ETD_Jump,JUMPGAD,IntuitionBase);
			relright -= ewidth;
			db_width -= ewidth;
		}

		if (((struct IntWindow *)(w))->extrabuttons & ETG_SNAPSHOT)
		{
			ewidth = addextragadget(w,is_gzz,dri,relright,SNAPSHOTIMAGE,((struct IntWindow *)w)->extrabuttonsid + ETD_Snapshot,SNAPSHOTGAD,IntuitionBase);
			relright -= ewidth;
			db_width -= ewidth;
		}

		if (((struct IntWindow *)(w))->extrabuttons & ETG_MUI)
		{
			ewidth = addextragadget(w,is_gzz,dri,relright,MUIIMAGE,((struct IntWindow *)w)->extrabuttonsid + ETD_MUI,MUIGAD,IntuitionBase);
			relright -= ewidth;
			db_width -= ewidth;
		}

		if (((struct IntWindow *)(w))->extrabuttons & ETG_POPUP)
		{
			ewidth = addextragadget(w,is_gzz,dri,relright,POPUPIMAGE,((struct IntWindow *)w)->extrabuttonsid + ETD_PopUp,POPUPGAD,IntuitionBase);
			relright -= ewidth;
			db_width -= ewidth;
		}

		if (w->Flags & WFLG_CLOSEGADGET)
		{
			struct TagItem close_tags[] =
			    {
				    {GA_Image	, 0		},
				    {GA_Left	, 0			},
				    {GA_Top		, 0			},
				    /*{GA_Width	, TITLEBAR_HEIGHT	},*/
				    {GA_Height	, TITLEBAR_HEIGHT	},
				    {GA_SysGadget	, TRUE			},
				    {GA_SysGType	, GTYP_CLOSE		},
				    {GA_TopBorder	, TRUE			},
				    {GA_GZZGadget	, is_gzz		},
				    {GA_RelVerify	, TRUE			},
				    {TAG_DONE				}
			    };
			struct TagItem image_tags[] =
			    {
				    //{IA_Width	, TITLEBAR_HEIGHT + 1	},
				    {IA_Height	, TITLEBAR_HEIGHT},
				    {SYSIA_Which	, CLOSEIMAGE		},
				    {SYSIA_DrawInfo , (IPTR)dri		},
				    {SYSIA_Size	, w->WScreen->Flags & SCREENHIRES ?
				     SYSISIZE_MEDRES : SYSISIZE_LOWRES},
				    {TAG_DONE				}
			    };
			Object *im;

			im = NewObjectA(NULL, SYSICLASS, image_tags);
			if (!im)
			{
				sysgads_ok = FALSE;
			}
			else
			{
				close_tags[0].ti_Data = (IPTR)im;

				SYSGAD(w, CLOSEGAD) = NewObjectA(NULL, BUTTONGCLASS, close_tags);

				if (!SYSGAD(w, CLOSEGAD))
				{
					DisposeObject(im);
					sysgads_ok = FALSE;
				}
				else
				{
					int width;
					GetAttr(GA_Width, SYSGAD(w, CLOSEGAD), (ULONG*)&width);

					db_left  += width;
					db_width -= width;
				}
			}
		}

		/* useful l8r */
		IW(w)->leftpos = db_left;
		IW(w)->rightpos = w->Width + relright - 1;

		if (w->Flags & WFLG_DRAGBAR)
		{

			struct TagItem dragbar_tags[] =
			{
				{GA_Left	, 0/*db_left*/},
				{GA_Top		, 0			},
				{GA_RelWidth	, 0/*db_width*/		},
				{GA_Height	, TITLEBAR_HEIGHT	},
				{GA_SysGadget	, TRUE			},
				{GA_SysGType	, GTYP_WDRAGGING	},
				{GA_TopBorder	, TRUE			},
				{GA_GZZGadget	, is_gzz		},
				{TAG_DONE				}
			};
			SYSGAD(w, DRAGBAR) = NewObjectA(NULL, BUTTONGCLASS, dragbar_tags);

			if (!SYSGAD(w, DRAGBAR))
				sysgads_ok = FALSE;

		}


		D(bug("Dragbar:  %p\n", SYSGAD(w, DRAGBAR ) ));
		D(bug("Depthgad: %p\n", SYSGAD(w, DEPTHGAD) ));
		D(bug("Zoomgad:  %p\n", SYSGAD(w, ZOOMGAD ) ));
		D(bug("Closegad: %p\n", SYSGAD(w, CLOSEGAD) ));
		D(bug("Sizegad:  %p\n", SYSGAD(w, SIZEGAD ) ));

		/* Don't need drawinfo anymore */
		FreeScreenDrawInfo(w->WScreen, dri);

		if (sysgads_ok)
		{
			int i;


			D(bug("Adding gadgets\n"));
			for (i = NUM_SYSGADS; --i >= 0; )
			{
				if (SYSGAD(w, i))
					AddGadget(w, (struct Gadget *)SYSGAD(w, i), 0);
			}

			ReturnBool("CreateWinSysGadgets", TRUE);

		} /* if (sysgads created) */

		KillWinSysGadgets(w, IntuitionBase);

	} /* if (got DrawInfo) */
	ReturnBool("CreateWinSysGadgets", FALSE);

}

/**********************************************************************************/

VOID KillWinSysGadgets(struct Window *w, struct IntuitionBase *IntuitionBase)
{
	/* Free system gadgets */
	UWORD i;

	for (i = 0; i < NUM_SYSGADS; i ++)
	{
		if (SYSGAD(w, i))
		{
			RemoveGadget( w, (struct Gadget *)SYSGAD(w, i));
			DisposeObject((Object *)((struct Gadget *)SYSGAD(w, i))->GadgetRender);
			DisposeObject( SYSGAD(w, i) );
			SYSGAD(w,i) = 0;
		}
	}
}

/**********************************************************************************/

void CreateScreenBar(struct Screen *scr, struct IntuitionBase *IntuitionBase)
{
	BOOL front = TRUE;
	ULONG moreflags = NULL;
	ULONG backdrop = LAYERBACKDROP;
	WORD ypos = 0;

#ifdef SKINS
	if (scr->Flags & SCREENQUIET || (IS(scr)->SpecialFlags & SF_InvisibleBar)) front = FALSE;
	if (IS(scr)->SpecialFlags & SF_AppearingBar)
	{
		backdrop = 0;
		ypos = - (scr->BarHeight + 1);
		moreflags |= LAYERONTOP;
	}
#else
	if (scr->Flags & SCREENQUIET) front = FALSE:
#endif

	if (!scr->BarLayer)
	{
		struct TagItem layertags[] =
		{
			{LA_BackfillHook, (IPTR)LAYERS_NOBACKFILL},
			{(IS(scr)->SpecialFlags & SF_AppearingBar) ? LA_Opacity : TAG_IGNORE, 0},
			{TAG_DONE}
		};

		if (front)
		{
			scr->BarLayer = CreateUpfrontLayerTagList(&scr->LayerInfo,
												   scr->RastPort.BitMap,
												   0,
												   ypos,
												   scr->Width - 1,
												   scr->BarHeight + ypos, /* 1 pixel heigher than scr->BarHeight */
												   LAYERSIMPLE | backdrop | moreflags,
												   (struct TagItem *)&layertags);
		}
		else
		{
			scr->BarLayer = CreateBehindLayerTagList(&scr->LayerInfo,
												   scr->RastPort.BitMap,
												   0,
												   ypos,
												   scr->Width - 1,
												   scr->BarHeight + ypos, /* 1 pixel heigher than scr->BarHeight */
												   LAYERSIMPLE | backdrop,
												   (struct TagItem *)&layertags);
		}

		if (scr->BarLayer)
		{
			SetFont(scr->BarLayer->rp, ((struct IntScreen *)scr)->DInfo.dri_Font);
			RenderScreenBar(scr, FALSE, IntuitionBase);
		}

	}
}

/**********************************************************************************/

void KillScreenBar(struct Screen *scr, struct IntuitionBase *IntuitionBase)
{
	if (scr->BarLayer)
	{
		DeleteLayer(0, scr->BarLayer);
		scr->BarLayer = FALSE;
	}

}

/**********************************************************************************/

#ifdef SKINS
//RenderScreenBar moved to morphos/mosmisc.c
#endif

#ifndef SKINS

void RenderScreenBar(struct Screen *scr, BOOL refresh, struct IntuitionBase *IntuitionBase)
{

	struct DrawInfo *dri = &((struct IntScreen *)scr)->DInfo;
	struct RastPort *rp;

	if (scr->BarLayer)
	{
#if USE_NEWDISPLAYBEEP
		BOOL beeping;
#else
#define	beeping 0
#endif

		rp = scr->BarLayer->rp;

		/* must lock GadgetLock to avoid deadlocks with ObtainGIRPort
		   when calling refreshgadget inside layer update state */
		LockLayerInfo(scr->BarLayer->LayerInfo);
		LOCKGADGET
		LockLayer(NULL,scr->BarLayer);

#if USE_NEWDISPLAYBEEP
		beeping = (scr->Flags & BEEPING) && GetBitMapAttr(rp->BitMap, BMA_DEPTH) > 8;
#endif

		if (refresh) BeginUpdate(scr->BarLayer);


		/* real BarHeight = scr->BarHeight + 1 !!! */


		SetDrMd(rp, JAM2);
		SetAPen(rp, dri->dri_Pens[beeping ? BARDETAILPEN : BARBLOCKPEN]);
		RectFill(rp, 0, 0, scr->Width - 1, scr->BarHeight - 1);

		SetAPen(rp, dri->dri_Pens[beeping ? BARDETAILPEN : BARTRIMPEN]);
		RectFill(rp, 0, scr->BarHeight, scr->Width - 1, scr->BarHeight);

		if (scr->Title)
		{
			CONST_STRPTR title = patchScreenTitle(scr->Title);

			SetAPen(rp, dri->dri_Pens[beeping ? BARBLOCKPEN: BARDETAILPEN]);
			SetBPen(rp, dri->dri_Pens[beeping ? BARDETAILPEN : BARBLOCKPEN]);

			Move(rp, scr->BarHBorder, scr->BarVBorder + rp->TxBaseline);

			Text(rp, title, strlen(title));
		}

		if (scr->FirstGadget)
		{
			RefreshBoopsiGadget(scr->FirstGadget, (struct Window *)scr, NULL, IntuitionBase);
		}

		if (refresh)
		{
			scr->BarLayer->Flags &= ~LAYERREFRESH;
			EndUpdate(scr->BarLayer, TRUE);
		}



		UnlockLayer(scr->BarLayer);
		UNLOCKGADGET
		UnlockLayerInfo(scr->BarLayer->LayerInfo);

	} /* if (scr->BarLayer) */

}

#endif

/**********************************************************************************/

void UpdateMouseCoords(struct Window *win)
{
	WORD scrmousex = win->WScreen->MouseX;
	WORD scrmousey = win->WScreen->MouseY;

	win->MouseX    = scrmousex - win->LeftEdge;
	win->MouseY    = scrmousey - win->TopEdge;

	/* stegerg: AmigaOS sets this even if window is not GZZ
	   so we do the same as they are handy also for non-GZZ
	   windows */

	win->GZZMouseX = scrmousex - (win->LeftEdge + win->BorderLeft);
	win->GZZMouseY = scrmousey - (win->TopEdge + win->BorderTop);
}

/**********************************************************************************/

/* subtract rectangle b from rectangle b. resulting rectangles will be put into
   destrectarray which must have place for at least 4 rectangles. Returns number
   of resulting rectangles */

#if 0  /* use <clib/macros.h> MAX/MIN macros */
#undef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))

#undef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

WORD SubtractRectFromRect(struct Rectangle *a, struct Rectangle *b, struct Rectangle *destrectarray)
{
	struct Rectangle	intersect;
	BOOL		intersecting = FALSE;
	WORD		numrects = 0;

	/* calc. intersection between a and b */

	if (a->MinX <= b->MaxX)
	{
		if (a->MinY <= b->MaxY)
		{
			if (a->MaxX >= b->MinX)
			{
				if (a->MaxY >= b->MinY)
				{
					intersect.MinX = MAX(a->MinX, b->MinX);
					intersect.MinY = MAX(a->MinY, b->MinY);
					intersect.MaxX = MIN(a->MaxX, b->MaxX);
					intersect.MaxY = MIN(a->MaxY, b->MaxY);

					intersecting = TRUE;
				}
			}
		}
	}

	if (!intersecting)
	{
		destrectarray[numrects++] = *a;

	} /* not intersecting */
	else
	{
		if (intersect.MinY > a->MinY) /* upper */
		{
			destrectarray->MinX = a->MinX;
			destrectarray->MinY = a->MinY;
			destrectarray->MaxX = a->MaxX;
			destrectarray->MaxY = intersect.MinY - 1;

			numrects++;
			destrectarray++;
		}

		if (intersect.MaxY < a->MaxY) /* lower */
		{
			destrectarray->MinX = a->MinX;
			destrectarray->MinY = intersect.MaxY + 1;
			destrectarray->MaxX = a->MaxX;
			destrectarray->MaxY = a->MaxY;

			numrects++;
			destrectarray++;
		}

		if (intersect.MinX > a->MinX) /* left */
		{
			destrectarray->MinX = a->MinX;
			destrectarray->MinY = intersect.MinY;
			destrectarray->MaxX = intersect.MinX - 1;
			destrectarray->MaxY = intersect.MaxY;

			numrects++;
			destrectarray++;
		}

		if (intersect.MaxX < a->MaxX) /* right */
		{
			destrectarray->MinX = intersect.MaxX + 1;
			destrectarray->MinY = intersect.MinY;
			destrectarray->MaxX = a->MaxX;
			destrectarray->MaxY = intersect.MaxY;

			numrects++;
			destrectarray++;
		}

	} /* intersecting */

	return numrects;

}

ULONG addextragadget(struct Window *w,BOOL is_gzz,struct DrawInfo *dri,LONG relright,ULONG imagetype,ULONG gadgetid,ULONG gadgettype,struct IntuitionBase *IntuitionBase)
{
	struct TagItem gadget_tags[] =
	    {
		    {GA_Image	, 0					},
		    {GA_ToggleSelect, FALSE			},
		    {GA_Top		, 0					},
		    {GA_Height	, TITLEBAR_HEIGHT	},
		    {GA_TopBorder	, TRUE			},
		    {GA_GZZGadget	, is_gzz		},
		    {GA_ID		, gadgetid			},
		    {GA_RelVerify, TRUE				},
		    {TAG_DONE						}
	    };
	struct TagItem image_tags[] =
	    {
#ifdef SKINS
	    	{TAG_IGNORE	, 0			},
#else
			{IA_Left	, -1			},
#endif
		    {IA_Height	, TITLEBAR_HEIGHT	},
		    {SYSIA_Which	, imagetype		},
		    {SYSIA_DrawInfo , (IPTR)dri		},
		    {SYSIA_Size	, w->WScreen->Flags & SCREENHIRES ?
		     SYSISIZE_MEDRES : SYSISIZE_LOWRES},
		    {TAG_DONE				}
	    };
	Object *im;

	if (gadgettype == LOCKGAD)
	gadget_tags[1].ti_Data = TRUE;

	im = NewObjectA(NULL, SYSICLASS, image_tags);
	if (im)
	{
		gadget_tags[0].ti_Data = (IPTR)im;

		SYSGAD(w, gadgettype) = NewObjectA(NULL, BUTTONGCLASS, gadget_tags);

		if (!SYSGAD(w, gadgettype))
		{
			DisposeObject(im);
		}
		else
		{
			int width;
			GetAttr(GA_Width, SYSGAD(w, gadgettype), (ULONG*)&width);

			--width;

			{
				struct TagItem gadtags[] =
				    {
					    {
						    GA_RelRight, relright - width
					    },
					    {TAG_DONE			    }
				    };


				SetAttrsA(SYSGAD(w, gadgettype), gadtags);
			}

			return width;
		}
	}

	return 0;

}

/**********************************************************************************/
